/*
 * SBuf.cc (C) 2008 Francesco Chemolli <kinkie@squid-cache.org>
 *
 * SQUID Web Proxy Cache          http://www.squid-cache.org/
 * ----------------------------------------------------------
 *
 *  Squid is the result of efforts by numerous individuals from
 *  the Internet community; see the CONTRIBUTORS file for full
 *  details.   Many organizations have provided support for Squid's
 *  development; see the SPONSORS file for full details.  Squid is
 *  Copyrighted (C) 2001 by the Regents of the University of
 *  California; see the COPYRIGHT file for full details.  Squid
 *  incorporates software developed and/or copyrighted by other
 *  sources; see the CREDITS file for full details.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
 */

#include "squid.h"
#include "base/RefCount.h"
#include "Debug.h"
#include "SBuf.h"
#include "SBufExceptions.h"
#include "OutOfBoundsException.h"
#include "util.h"

#if HAVE_STRING_H
#include <string.h>
#endif

#if HAVE_SSTREAM
#include <sstream>
#endif

#if HAVE_IOSTREAM
#include <iostream>
#endif

#ifdef VA_COPY
#undef VA_COPY
#endif
#if defined HAVE_VA_COPY
#define VA_COPY va_copy
#elif defined HAVE___VA_COPY
#define VA_COPY __va_copy
#endif

InstanceIdDefinitions(SBuf, "SBuf");

SBufStats SBuf::stats;
MemBlob::Pointer SBuf::StorePrototype = 0;

SBufStats::SBufStats()
        : alloc(0), allocCopy(0), allocFromString(0), allocFromCString(0),
        assignFast(0), clear(0), append(0), toStream(0), setChar(0),
        getChar(0), compareSlow(0), compareFast(0), copyOut(0),
        rawAccess(0), chop(0), trim(0), find(0), scanf(0),
        caseChange(0), cowFast(0), cowSlow(0), live(0)
{}

SBufStats&
SBufStats::operator +=(const SBufStats& ss)
{
    alloc += ss.alloc;
    allocCopy += ss.allocCopy;
    allocFromString += ss.allocFromString;
    allocFromCString += ss.allocFromCString;
    assignFast += ss.assignFast;
    clear += ss.clear;
    append += ss.append;
    toStream += ss.toStream;
    setChar += ss.setChar;
    getChar += ss.getChar;
    compareSlow += ss.compareSlow;
    compareFast += ss.compareFast;
    copyOut += ss.copyOut;
    rawAccess += ss.rawAccess;
    chop += ss.chop;
    trim += ss.trim;
    find += ss.find;
    scanf += ss.scanf;
    caseChange += ss.caseChange;
    cowFast += ss.cowFast;
    cowSlow += ss.cowSlow;
    live += ss.live;

    return *this;
}

SBuf::SBuf()
        : store_(GetStorePrototype()), off_(0), len_(0)
{
    debugs(24, 8, id << " created");
    ++stats.alloc;
    ++stats.live;
}

SBuf::SBuf(const SBuf &S)
        : store_(S.store_), off_(S.off_), len_(S.len_)
{
    debugs(24, 8, id << " created from id " << S.id);
    ++stats.alloc;
    ++stats.allocCopy;
    ++stats.live;
}

SBuf::SBuf(const String &S)
        : store_(GetStorePrototype()), off_(0), len_(0)
{
    debugs(24, 8, id << " created from string");
    assign(S.rawBuf(), 0, S.size());
    ++stats.alloc;
    ++stats.allocFromString;
    ++stats.live;
}

SBuf::SBuf(const char *S, size_type pos, size_type n)
        : store_(GetStorePrototype()), off_(0), len_(0)
{
    append(S,pos,n);
    ++stats.alloc;
    ++stats.allocFromCString;
    ++stats.live;
}

SBuf::~SBuf()
{
    debugs(24, 8, id << " destructed");
    --stats.live;
}

SBuf&
SBuf::assign(const SBuf &S)
{
    debugs(24, 7, "assigning " << id << " from " <<  S.id);
    if (&S == this) //assignment to self. Noop.
        return *this;
    ++stats.assignFast;
    store_ = S.store_;
    off_ = S.off_;
    len_ = S.len_;
    return *this;
}

SBuf&
SBuf::assign(const char *S, size_type pos, size_type n)
{
    debugs(24, 6, id << " from c-string, pos=" << pos << "n=" << n << ")");
    clear();
    return append(S, pos, n);
}

void
SBuf::reserve(size_type minSize)
{
    debugs(24, 7, "reserving " << minSize << " for " << id);
    // we're not concerned about RefCounts here,
    // the store knows the last-used portion. If
    // it's available, we're effectively claiming ownership
    // of it. If it's not, we need to go away (realloc)
    if (store_->canAppend(off_+len_, minSize)) {
        debugs(24, 7, "not growing");
        return;
    }
    // TODO: we may try to memmove before realloc'ing in order to avoid
    //   one allocation operation, if we're the sole owners of a MemBlob.
    //   Maybe some heuristic on off_ and length()?
    reAlloc(estimateCapacity(minSize+length()));
}

void
SBuf::clear()
{
#if 0
    //enabling this code path, the store will be freed and reinitialized
    store_ = GetStorePrototype(); //uncomment to actually free storage upon clear()
#else
    //enabling this code path, we try to release the store without deallocating it.
    // will be lazily reallocated if needed.
    if (store_->LockCount() == 1)
        store_->clear();
#endif
    len_ = 0;
    off_ = 0;
    ++stats.clear;
}

SBuf&
SBuf::append(const SBuf &S)
{
    return append(S.buf(), 0, S.length());
}

SBuf&
SBuf::append(const char * S, size_type pos, size_type n)
{
    Must(pos == npos || pos >= 0);
    Must(n == npos || n >= 0);

    if (S == NULL)
        return *this;
    if (n == npos)
        n = strlen(S)-pos;

    debugs(24, 7, "from c-string to id " << id);

    reserve(n);
    const char *actual_start = S+pos;
    store_->append(actual_start, n);
    len_ += n;
    ++stats.append;
    return *this;
}

SBuf&
SBuf::append(const std::string &str, SBuf::size_type pos, SBuf::size_type n)
{
    return append(str.data(), pos, n);
}

SBuf&
SBuf::assign(const std::string &str, size_type pos, size_type n)
{
    clear();
    return append(str, pos, n);
}

SBuf&
SBuf::Printf(const char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    clear();
    vappendf(fmt, args);
    va_end(args);
    return *this;
}

SBuf&
SBuf::appendf(const char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    vappendf(fmt, args);
    va_end(args);
    return *this;
}

SBuf&
SBuf::vappendf(const char *fmt, va_list vargs)
{
#ifdef VA_COPY
    va_list ap;
#endif
    int sz = 0;

    Must(fmt != NULL);

    //we can assume that we'll need to append at least strlen(fmt) bytes,
    //times 1.2 for instance...
    reserve(strlen(fmt)*2);

    while (length() <= maxSize) {
#ifdef VA_COPY
        /* Fix of bug 753r. The value of vargs is undefined
         * after vsnprintf() returns. Make a copy of vargs
         * in case we loop around and call vsnprintf() again.
         */
        VA_COPY(ap, vargs);
        sz = vsnprintf(bufEnd(), store_->spaceSize(), fmt, ap);
        va_end(ap);
#else /* VA_COPY */
        sz = vsnprintf(bufEnd(), store_->spaceSize(), fmt, vargs);
#endif /* VA_COPY*/
        /* check for possible overflow */
        /* snprintf on Linux returns -1 on overflows */
        /* snprintf on FreeBSD returns at least free_space on overflows */

        if (sz < 0 || sz >= (int)store_->spaceSize())
            reserve(length()+sz*2); // TODO: tune heuristics
        else
            break;
    }

    len_ += sz;
    // TODO: this does NOT belong here, but to class-init or autoconf
    /* on Linux and FreeBSD, '\0' is not counted in return value */
    /* on XXX it might be counted */
    /* check that '\0' is appended and not counted */

    if (operator[](len_-1) == 0) {
        --sz;
        --len_;
    }

    store_->size += sz;
    ++stats.append;

    return *this;
}

std::ostream&
SBuf::print(std::ostream &os) const
{
    os.write(buf(), length());
    ++stats.toStream;
    return os;
}

std::ostream&
SBuf::dump(std::ostream &os) const
{
    os << id
    << ": ";
    store_->dump(os);
    os << ",offset:" << off_
    << ",len:" << len_
    << ") : '";
    print(os);
    os << std::endl;
    return os;
}

void
SBuf::setAt(SBuf::size_type pos, char toset)
{
    checkAccessBounds(pos);
    cow();
    store_->mem[off_+pos] = toset;
    ++stats.setChar;
}

int
SBuf::compare(const SBuf &S, SBufCaseSensitive isCaseSensitive, size_type n) const
{
    int rv;
    size_type sz = min(S.length(), length());
    if (n != npos)
        sz = min(n, sz);
    ++stats.compareSlow;
    if (isCaseSensitive == caseSensitive)
        rv = strncmp(buf(), S.buf(), sz);
    else
        rv = strncasecmp(buf(), S.buf(), sz);
    if (rv != 0)
        return rv;
    //if max-length was specified, it was hit, and the tail is ignored.
    if (n != npos)
        return 0;
    //first sz bytes equal. longest string "wins"
    return commonCompareChecksPost(S);
}

bool
SBuf::startsWith(const SBuf &S, SBufCaseSensitive isCaseSensitive) const
{
    if (length() < S.length()) {
        ++stats.compareFast;
        return false;
    }
    return (compare(S, isCaseSensitive, S.length()) == 0);
}

bool
SBuf::operator ==(const SBuf & S) const
{
    debugs(24, 8, id << " == " << S.id);
    if (length() != S.length()) {
        debugs(24, 8, "no, different lengths");
        ++stats.compareFast;
        return false; //shortcut: must be equal length
    }
    if (store_ == S.store_ && off_ == S.off_) {
        debugs(24, 8, "yes, same length and backing store");
        ++stats.compareFast;
        return true;  //shortcut: same store, offset and length
    }
    ++stats.compareSlow;
    const bool rv = (0 == strncmp(buf(), S.buf(), length()));
    debugs(24, 8, "returning " << rv);
    return rv;
}

bool
SBuf::operator !=(const SBuf & S) const
{
    return !(*this == S);
}

SBuf
SBuf::consume(SBuf::size_type n)
{
    if (n == npos)
        n = length();
    else
        n = min(n, length());
    SBuf rv(substr(0, n));
    chop(n);
    return rv;
}

const
SBufStats& SBuf::GetStats()
{
    return stats;
}

SBuf::size_type
SBuf::copy(char *dest, SBuf::size_type n) const
{
    Must(n >= 0);

    SBuf::size_type toexport = length();
    if (toexport > n)
        toexport = n;

    memcpy(dest, buf(), toexport);

    ++stats.copyOut;
    return toexport;
}

const char*
SBuf::rawContent() const
{
    ++stats.rawAccess;
    return buf();
}

char *
SBuf::rawSpace(size_type minSize)
{
    cow(minSize); //passing default npos is handled by cow
    ++stats.rawAccess;
    return bufEnd();
}

void
SBuf::forceSize(SBuf::size_type newSize)
{
    Must(store_->LockCount() == 1);
    len_ = newSize;
    store_->size = newSize;
}

const char*
SBuf::c_str()
{
    ++stats.rawAccess;
    /* null-terminate the current buffer, by hand-appending a \0 at its tail but
     * without increasing its length. May COW, the side-effect is to guarantee that
     * the MemBlob's tail is availabe for us to use */
    reserve(1);
    *(bufEnd()) = '\0';
    ++store_->size;
    ++stats.setChar;
    return buf();
}

SBuf&
SBuf::chop(SBuf::size_type pos, SBuf::size_type n)
{
    Must(pos >= 0);
    Must(n == npos || n >= 0);
    /*
     * TODO: possible optimization: if the SBuf is at the tail of the
     * MemBlob we could decrease the MemBlob tail-pointer so that a subsequent
     * append will reuse the freed space.
     */
    if (pos > length() || n == 0) {
        clear();
        return *this;
    }
    if (n == npos || (pos+n) > length())
        n = length()-pos;
    ++stats.chop;
    off_ += pos;
    len_ = n;
    return *this;
}

SBuf&
SBuf::trim(const SBuf &toRemove, bool atBeginning, bool atEnd)
{
    ++stats.trim;
    if (atEnd) {
        const char *p = bufEnd()-1;
        while (!isEmpty() && memchr(toRemove.buf(), *p, toRemove.length()) != NULL) {
            //current end-of-buf is in the searched set
            --len_;
            --p;
        }
    }
    if (atBeginning) {
        const char *p = buf();
        while (!isEmpty() && memchr(toRemove.buf(), *p, toRemove.length()) != NULL) {
            --len_;
            ++off_;
            ++p;
        }
    }
    if (isEmpty())
        clear();
    return *this;
}

SBuf
SBuf::substr(SBuf::size_type pos, SBuf::size_type n) const
{
    SBuf rv(*this);
    rv.chop(pos, n); //stats handled by callee
    return rv;
}

SBuf::size_type
SBuf::find(char c, SBuf::size_type startPos) const
{
    if (startPos == npos)
        startPos = length();
    startPos = min(startPos, length());
    ++stats.find;
    void *i = memchr(buf()+startPos, (int)c, length()-startPos);
    if (i == NULL)
        return SBuf::npos;
    return ((const char *)i-buf());
}

SBuf::size_type
SBuf::find(const SBuf &str, size_type startPos) const
{
    if (startPos == npos)
        startPos = length();
    else
        startPos = min(startPos, length());
    char *begin = buf()+startPos, *end = buf()+length(), *tmp;
    char needle = str[0];

    debugs(24, 7, "looking for " << str << "starting at " << startPos <<
                    " in id " << id);
    ++stats.find;
    while (end > begin) {
        debugs(24, 8, " begin=" << (void *) begin <<
               ", end=" << (void*) end );
        tmp = static_cast<char *>(memchr(begin, needle, end-begin));
        if (tmp == NULL) { //first char not found
            debugs(24, 8 , "First byte not found");
            return SBuf::npos;
        }
        if (tmp+str.length() > end) { //not enough chars to match the whole needle
            debugs(24, 8 , "needle too big for haystack");
            return SBuf::npos;
        }
        //we are guarranteed not to go out-of-bounds with the find
        if (0 == memcmp(str.buf(), tmp, str.length())) {
            //found
            debugs(24, 8, "Found at " << (tmp-buf()));
            return (tmp-buf());
        }
        begin = tmp+1;
    }
    debugs(24, 8, "not found");
    return SBuf::npos;
}

SBuf::size_type
SBuf::rfind(const SBuf &str, SBuf::size_type endpos) const
{
    char *begin = buf(), *end, *tmp;
    if (endpos == npos) {
        end = buf()+length();
    } else {
        endpos = min(endpos, length());
        end = buf()+endpos;
    }
    char needle = str[0];

    debugs(24, 7, "looking for " << str << "ending at " << endpos <<
                    " in id " << id);
    ++stats.find;
    while (end > begin) {
        debugs(24, 8, " begin=" << (void*) begin <<
               ", end=" << (void*) end );
        tmp = static_cast<char *>(memrchr(begin, needle, end-begin));
        if (tmp == NULL) { //first char not found
            debugs(24, 8, "First byte not found");
            return SBuf::npos;
        }
        if (buf()+length() < tmp+str.length()) { //not enough chars to match the whole needle
            debugs(24, 8, "needle too big for haystack");
            end = tmp-1;
            continue;
        }
        //we are guarranteed not to go out-of-bounds with the find
        if (0 == memcmp(str.buf(), tmp, str.length())) {
            //found
            debugs(24, 8, "Found at " << (tmp-buf()));
            return (tmp-buf());
        }
        end = tmp-1;
    }
    debugs(24, 8, "not found");
    return SBuf::npos;
}

SBuf::size_type
SBuf::rfind(char c, SBuf::size_type endpos) const
{
    debugs(24, 7, "looking for char " << c << "ending at " << endpos <<
                    " in id " << id);
    size_type searchlen;
    if (endpos == npos)
        searchlen = length();
    else
        searchlen = min(endpos, length());
    ++stats.find;
    void *i = memrchr(buf(), (int)c, searchlen);
    if (i == NULL) {
        debugs(24, 7, "not found");
        return SBuf::npos;
    }
    const size_type rv = (static_cast<const char *>(i)-buf());
    debugs(24, 7, "found at position " << rv);
    return rv;
}

SBuf::size_type
SBuf::find_first_of(const SBuf &set, size_type startPos) const
{
    Must(startPos == npos || startPos >= 0);
    Must(set.length() > 0);
    debugs(24, 7, "any of '" << set << "' " << " in id " << id);
    if (startPos == npos || startPos > length())
        return npos;
    startPos = min(startPos, length());
    char *cur = buf()+startPos, *end = bufEnd();
    ++stats.find;
    while (cur < end) {
        char *s1 = set.buf(), *s2 = set.buf()+set.length();
        while (s1 < s2) {
            if (*s1 == *cur) {
                debugs(24, 7, "found at position " << cur-buf());
                return (cur-buf());
            }
            ++s1;
        }
        ++cur;
    }
    debugs(24, 7, "not found");
    return npos;
}

/*
 * TODO: borrow a sscanf implementation from Linux or similar?
 * we'd really need a vsnscanf(3)... ? As an alternative, a
 * light-regexp-like domain-specific syntax might be an idea.
 */
int
SBuf::scanf(const char *format, ...)
{
    va_list arg;
    int rv;
    ++stats.scanf;
    va_start(arg, format);
    rv = vsscanf(c_str(), format, arg);
    va_end(arg);
    return rv;
}

std::ostream &
operator <<(std::ostream& os, const SBuf& S)
{
    return S.print(os);
}

std::ostream &
SBufStats::dump(std::ostream& os) const
{
    MemBlobStats ststats = MemBlob::GetStats();
    os <<
    "SBuf stats:\nnumber of allocations: " << alloc <<
    "\ncopy-allocations: " << allocCopy <<
    "\ncopy-allocations from SquidString: " << allocFromString <<
    "\ncopy-allocations from C String: " << allocFromCString <<
    "\nlive references: " << live <<
    "\nno-copy assignments: " << assignFast <<
    "\nclearing operations: " << clear <<
    "\nappend operations: " << append <<
    "\ndump-to-ostream: " << toStream <<
    "\nset-char: " << setChar <<
    "\nget-char: " << getChar <<
    "\ncomparisons with data-scan: " << compareSlow <<
    "\ncomparisons not requiring data-scan: " << compareFast <<
    "\ncopy-out ops: " << copyOut <<
    "\nraw access to memory: " << rawAccess <<
    "\nchop operations: " << chop <<
    "\ntrim operations: " << trim <<
    "\nfind: " << find <<
    "\nscanf: " << scanf <<
    "\ncase-change ops: " << caseChange <<
    "\nCOW not actually requiring a copy: " << cowFast <<
    "\nCOW: " << cowSlow <<
    "\naverage store share factor: " <<
    (ststats.live != 0 ? static_cast<float>(live)/ststats.live : 0) <<
    std::endl;
    return os;
}

SBuf
SBuf::toLower() const
{
    debugs(24, 8, "\"" << *this << "\"");
    SBuf rv(*this);
    for (size_type j = 0; j < length(); ++j) {
        int c = (*this)[j];
        if (isupper(c))
            rv.setAt(j, tolower(c)); //will cow() if needed
    }
    debugs(24, 8, "result: \"" << *this << "\"");
    ++stats.caseChange;
    return rv;
}

SBuf
SBuf::toUpper() const
{
    debugs(24, 8, "\"" << *this << "\"");
    SBuf rv(*this);
    for (size_type j = 0; j < length(); ++j) {
        int c = (*this)[j];
        if (islower(c))
            rv.setAt(j, toupper(c)); //will cow() if needed
    }
    debugs(24, 8, "result: \"" << *this << "\"");
    ++stats.caseChange;
    return rv;
}

/**
 * checks whether the requested 'pos' is within the bounds of the SBuf
 * \throw OutOfBoundsException if access is out of bounds
 */
void
SBuf::checkAccessBounds(SBuf::size_type pos) const
{
    if (pos < 0)
        throw OutOfBoundsException(*this, pos, __FILE__, __LINE__);
    if (pos > length())
        throw OutOfBoundsException(*this, pos, __FILE__, __LINE__);
}

String
SBuf::toString() const
{
    String rv;
    rv.limitInit(buf(), length());
    ++stats.copyOut;
    return rv;
}

#if !_USE_INLINE_
#include "SBuf.cci"
#endif

