/*
 * SBuTokenizer.h (C) 2008 Francesco Chemollli <kinkie@squid-cache.org>
 *
 * SQUID Web Proxy Cache          http://www.squid-cache.org/
 * ----------------------------------------------------------
 *
 *  Squid is the result of efforts by numerous individuals from
 *  the Internet community; see the CONTRIBUTORS file for full
 *  details.   Many organizations have provided support for Squid's
 *  development; see the SPONSORS file for full details.  Squid is
 *  Copyrighted (C) 2001 by the Regents of the University of
 *  California; see the COPYRIGHT file for full details.  Squid
 *  incorporates software developed and/or copyrighted by other
 *  sources; see the CREDITS file for full details.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
 */

#ifndef SQUID_SBUFTOKENIZER_H
#define SQUID_SBUFTOKENIZER_H

#include "SBuf.h"

/**
 * A tokenizer for SBuf
 * example:
 * \code
 * SBuf toparse, sep;
 * toparse = ...; sep = ...;
 * for (SBufTokenizer st(toparse,sep);!st.atEnd();st.next()) {
 *     SBuf s = st.token();
 *     //...
 * }
 * \endcode
 */
class SBufTokenizer
{
public:
    /** Constructor
     *
     * create a new SBufTokenizer to check over the contents of toparse,
     * using any of the characters of delim as a token delimiter
     * \param toparse the SBuf to be tokenized
     * \param delim the token separator. It's a SBuf which is interpreted
     *   as a bag of chars, whose each is a possible token delimiter.
     */
    SBufTokenizer(const SBuf &toparse, const SBuf &delim)
    {
        reset(toparse,delim);
    }

    /** reset the Tokenizer
     *
     * reset the SBufTokenizer with a new SBuf to be parsed and a new
     * delimiter set.
     * \param toparse the SBuf to be tokenized
     * \param delim the token separator. It's a SBuf which is interpreted
     */
    void reset(const SBuf &toparse, const SBuf &delim);

    /** set a new delimiter set
     *
     * \param newdelim the token separator.
     */
    void setDelimiter(const SBuf &newdelim)
    {
        _delim = newdelim;
    }

    /** retrieve the current delimiter set
     */
    SBuf getDelimiter() const
    {
        return _delim;
    }

    /** Extract the next Token
     *
     * strtok()-equivalent.
     * Will extract a token from the beginning of the SBuf, containing
     * all chars up to the first occurrence of any of the chars in
     * <i>delim</i> or end-of-SBuf if no delimiter can be found.
     * The tokenized SBuf is modified, by removing its head
     * the to-be-returned token AND the delimiter.
     */
    SBufTokenizer& next();

    /** get last delimiter found
     *
     ** returns the delimiter which terminated the last-matched token
     * If no token() has not been extracted or we are at the end of the
     * SBuf to be tokenized, the result is undefined
     */
    char separator() const
    {
        return _lastDelim;
    }

    /** Move to the next token, specify new delimiter
     *
     * Move to the next token, and a new delimiter to parse it
     */
    SBufTokenizer &next(const SBuf &newdelim)
    {
        setDelimiter(newdelim);
        return next();
    }

    /** Extract a token
     *
     * Extract a token from the SBuf. May be called repeatedly, in that case
     * will return each time a fresh copy of the same token.
     */
    SBuf token() const
    {
        return _lastToken;
    }

    /** check whether we are done
     *
     * \retval true if we are at the SBuf to be parsed.
     */
    bool atEnd() const;

private:
    SBuf _toParse;  /// the still-unparsed part of the buffer to parse
    SBuf _lastToken;/// the last returned token
    char _lastDelim;/// the delimiter last matched
    SBuf _delim;    /// a SBuf containing the delimiter characters

    SBuf::size_type pbrk() const
    {
        return _toParse.find_first_of(_delim);
    }
};

#endif
