#include "squid.h"
#include "Mem.h"
#include "SBuf.h"
#include "SBufList.h"
#include "SBufStream.h"
#include "SBufTokenizer.h"
#include "SBufUtil.h"
#include "SquidString.h"
#include "testSBuf.h"

#include <iostream>
#include <stdexcept>

CPPUNIT_TEST_SUITE_REGISTRATION( testSBuf );

/* let this test link sanely */
#include "event.h"
#include "MemObject.h"
void
eventAdd(const char *name, EVH * func, void *arg, double when, int, bool cbdata)
{}
int64_t
MemObject::endOffset() const
{ return 0; }
/* end of stubs */

// test string
static char fox[]="The quick brown fox jumped over the lazy dog";
static char fox1[]="The quick brown fox ";
static char fox2[]="jumped over the lazy dog";

// TEST: globals variables (default/empty and with contents) are
//  created outside and before any unit tests and memory subsystem
//  initialization. Check for correct constructor operation.
SBuf empty_sbuf;
SBuf literal("The quick brown fox jumped over the lazy dog");

void
testSBuf::testSBufConstructDestruct()
{
    /* NOTE: Do not initialize memory here because we need
     * to test correct operation before and after Mem::Init
     */

    // XXX: partial demo below of how to do constructor unit-test. use scope to ensure each test
    // is working on local-scope variables constructed fresh for the test, and destructed when
    // scope exists. use nested scopes to test destructor affects on copied data (MemBlob etc)

    // TEST: default constructor (implicit destructor non-crash test)
    //  test accessors on empty SBuf.
    {
        SBuf s1;
        CPPUNIT_ASSERT_EQUAL(s1.length(),0);
        CPPUNIT_ASSERT_EQUAL(s1,SBuf(""));
        CPPUNIT_ASSERT_EQUAL(s1,empty_sbuf);
        CPPUNIT_ASSERT(0==strcmp("",s1.c_str()));
    }

    // TEST: copy-construct NULL string (implicit destructor non-crash test)
    {
        SBuf s1(NULL);
        CPPUNIT_ASSERT_EQUAL(s1.length(),0);
        CPPUNIT_ASSERT_EQUAL(s1,SBuf(""));
        CPPUNIT_ASSERT_EQUAL(s1,empty_sbuf);
        CPPUNIT_ASSERT(0==strcmp("",s1.c_str()));
    }

    // TEST: copy-construct empty string (implicit destructor non-crash test)
    {
        SBuf s1("");
        CPPUNIT_ASSERT_EQUAL(s1.length(),0);
        CPPUNIT_ASSERT_EQUAL(s1,SBuf(""));
        CPPUNIT_ASSERT_EQUAL(s1,empty_sbuf);
        CPPUNIT_ASSERT(0==strcmp("",s1.c_str()));
    }

    // TEST: copy-construct from a SBuf
    {
        SBuf s1(empty_sbuf);
        CPPUNIT_ASSERT_EQUAL(s1.length(),0);
        CPPUNIT_ASSERT_EQUAL(s1,SBuf(""));
        CPPUNIT_ASSERT_EQUAL(s1,empty_sbuf);
        CPPUNIT_ASSERT(0==strcmp("",s1.c_str()));

        SBuf s5(literal);
        CPPUNIT_ASSERT_EQUAL(s5,literal);
        SBuf s6(fox);
        CPPUNIT_ASSERT_EQUAL(s6,literal);
        // XXX: other state checks. expected result of calling any state accessor on s4 ?
    }

    // TEST: check that COW doesn't happen upon copy-construction
    {
        SBuf s1(empty_sbuf), s2(s1);
        CPPUNIT_ASSERT_EQUAL(s1.rawContent(), s2.rawContent());
        SBuf s3(literal), s4(literal);
        CPPUNIT_ASSERT_EQUAL(s3.rawContent(), s4.rawContent());
    }

    // TEST: sub-string copy
    {
        SBuf s1=SBuf(fox,4), s2(fox);
        SBuf s3=s2.substr(4,s2.length()); //n is out-of-bounds
        CPPUNIT_ASSERT_EQUAL(s1,s3);
        SBuf s4=SBuf(fox,0,4);
        s3=s2.substr(0,4);
        CPPUNIT_ASSERT_EQUAL(s4,s3);
    }

    // TEST: go via SquidString adapters.
    {
        String str(fox);
        SBuf s1(str);
        CPPUNIT_ASSERT_EQUAL(s1,literal);
    }
}

void
testSBuf::testSBufConstructDestructAfterMemInit()
{
    Mem::Init();
    testSBufConstructDestruct();

// XXX: or perhapse ...
// repeat all of the tests inside testSBufConstructDestructBeforeMemInit()
// with additional checks on Mem usage stats after each operation ??
}

void
testSBuf::testEqualityTest()
{
    SBuf s1(fox),s2(fox);
    CPPUNIT_ASSERT_EQUAL(s1,s1); //self-equality
    CPPUNIT_ASSERT_EQUAL(s1,s2); //same contents
    s2.assign("The quick brown fox jumped over the lazy doe");
    CPPUNIT_ASSERT(!(s1 == s2)); //same length, different contents
    s2.assign("foo");
    CPPUNIT_ASSERT(!(s1 == s2)); //different length and contents
    CPPUNIT_ASSERT(s1 != s2);    //while we're ready, let's test inequality
    s2.clear();
    CPPUNIT_ASSERT(!(s1 == s2)); //null and not-null
    CPPUNIT_ASSERT(s1 != s2);    //while we're ready, let's test inequality
    s1.clear();
    CPPUNIT_ASSERT_EQUAL(s1,s2); //null and null
}

void
testSBuf::testAppendSBuf()
{
    SBuf s1(fox1),s2(fox2);
    s1.append(s2);
    CPPUNIT_ASSERT_EQUAL(s1,literal);
}

void
testSBuf::testPrintf()
{
    SBuf s1,s2;
    s1.Printf("%s:%d:%03.3f","fox",10,12345.67);
    s2.assign("fox:10:12345.670");
    CPPUNIT_ASSERT_EQUAL(s1,s2);
}

void
testSBuf::testAppendCString()
{
    SBuf s1(fox1);
    s1.append(fox2);
    CPPUNIT_ASSERT_EQUAL(s1,literal);
}

void
testSBuf::testAppendStdString()
{
    SBuf s1(fox1);
    std::string str(fox2);
    s1.append(str);
    CPPUNIT_ASSERT_EQUAL(s1,literal);
}

void
testSBuf::testAppendf()
{
    SBuf s1,s2;
    s1.appendf("%s:%d:%03.2f",fox,1234,1234.56);
    s2.assign("The quick brown fox jumped over the lazy dog:1234:1234.56");
    CPPUNIT_ASSERT_EQUAL(s1,s2);
}

void
testSBuf::testDumpStats()
{
    SBuf::GetStats().dump(std::cout);
    MemBlob::GetStats().dump(std::cout);
    std::cout << "sizeof(SBuf): " << sizeof(SBuf) << std::endl;
    std::cout << "sizeof(MemBlob): " << sizeof(MemBlob) << std::endl;
}

void
testSBuf::testSubscriptOp()
{
    SBuf chg(literal);
    CPPUNIT_ASSERT_EQUAL(chg[5],'u');
    chg.setAt(5,'e');
    CPPUNIT_ASSERT_EQUAL(literal[5],'u');
    CPPUNIT_ASSERT_EQUAL(chg[5],'e');
//    std::cout << chg << std::endl << empty_sbuf << std::endl ;
}

// note: can't use cppunit's CPPUNIT_TEST_EXCEPTION because TextException asserts, and
// so the test can't be properly completed.
void
testSBuf::testSubscriptOpFail()
{
    char c;
    c=literal.at(1234); //out of bounds
    //notreached
    std::cout << c << std::endl;
}

static int sign(int v)
{
    if (v < 0)
        return -1;
    if (v>0)
        return 1;
    return 0;
}

void
testSBuf::testComparisons()
{
    //same length
    SBuf s1("foo"),s2("foe");
    CPPUNIT_ASSERT(s1.compare(s2)>0);
    CPPUNIT_ASSERT(s1.compare(s2,caseInsensitive,2)==0);
    CPPUNIT_ASSERT(s1 > s2);
    CPPUNIT_ASSERT(s2 < s1);
    CPPUNIT_ASSERT_EQUAL(sign(s1.compare(s2)),sign(strcmp(s1.c_str(),s2.c_str())));
    //different lengths
    s1.assign("foo");
    s2.assign("foof");
    CPPUNIT_ASSERT(s1.compare(s2)<0);
    CPPUNIT_ASSERT_EQUAL(sign(s1.compare(s2)),sign(strcmp(s1.c_str(),s2.c_str())));
    CPPUNIT_ASSERT(s1 < s2);
}

void
testSBuf::testConsume()
{
    SBuf s1(literal),s2,s3;
    s2=s1.consume(4);
    s3.assign("The ");
    CPPUNIT_ASSERT_EQUAL(s2,s3);
    s3.assign("quick brown fox jumped over the lazy dog");
    CPPUNIT_ASSERT_EQUAL(s1,s3);
    s1.consume(40);
    CPPUNIT_ASSERT_EQUAL(s1,SBuf());
}

void
testSBuf::testRawContent()
{
    SBuf s1(literal);
    SBuf s2(s1);
    s2.append("foo");
    const char *foo;
    foo = s1.rawContent();
    CPPUNIT_ASSERT(strncmp(fox,foo,s1.length())==0);
    foo = s1.c_str();
    CPPUNIT_ASSERT(!strcmp(fox,foo));
}

void
testSBuf::testRawSpace()
{
    SBuf s1(literal);
    SBuf s2(fox1);
    char *rb=s2.rawSpace(strlen(fox2)+1);
    strcat(rb,fox2);
    CPPUNIT_ASSERT_EQUAL(s1,s2);
}

void
testSBuf::testChop()
{
    SBuf s1(literal),s2;
    s1.chop(4,5);
    s2.assign("quick");
    CPPUNIT_ASSERT_EQUAL(s1,s2);
    s1=literal;
    s2.clear();
    s1.chop(5,0);
    CPPUNIT_ASSERT_EQUAL(s1,s2);
}

void
testSBuf::testChomp()
{
    SBuf s1("complete string");
    SBuf s2(s1);
    s2.trim(SBuf(" ,"));
    CPPUNIT_ASSERT_EQUAL(s1,s2);
    s2.assign(" complete string ,");
    s2.trim(SBuf(" ,"));
    CPPUNIT_ASSERT_EQUAL(s1,s2);
    s1.assign(", complete string ,");
    s2=s1;
    s2.trim(SBuf(" "));
    CPPUNIT_ASSERT_EQUAL(s1,s2);
}

void
testSBuf::testSubstr()
{
    SBuf s1(literal),s2,s3;
    s2=s1.substr(4,5);
    s3.assign("quick");
    CPPUNIT_ASSERT_EQUAL(s2,s3);
    s1.chop(4,5);
    CPPUNIT_ASSERT_EQUAL(s1,s2);
}

void
testSBuf::testFindChar()
{
    SBuf s1(literal);
    SBuf::size_type idx;
    idx=s1.find(' ');
    CPPUNIT_ASSERT(idx == 3);
    CPPUNIT_ASSERT(s1[idx]==' ');
    s1.consume(idx+1);
    idx=s1.find(' ');
    CPPUNIT_ASSERT(idx == 5);
    idx=s1.find('\020'); //fails
    CPPUNIT_ASSERT(idx==SBuf::npos);
    idx=s1.find('q',5);
    CPPUNIT_ASSERT(idx==SBuf::npos);
    idx=s1.rfind('g',s1.length()-3);
    CPPUNIT_ASSERT(idx==SBuf::npos);
}

void
testSBuf::testFindSBuf()
{
    SBuf s1(literal),s2("fox"),s3("foobar"),s4("foe"), s5("dog");
    //very long string matching at the end
    s5.append(literal).append(literal).append(literal).append(literal);
    SBuf::size_type idx;
    idx=s1.find(s2);
    CPPUNIT_ASSERT(idx == 16);
    idx=s1.find(s3); //fails
    CPPUNIT_ASSERT(idx==SBuf::npos);
    idx=s1.find(s4); //fails
    CPPUNIT_ASSERT(idx==SBuf::npos);
    idx=s1.find(s5);
    CPPUNIT_ASSERT(idx==SBuf::npos);
    s5="g"; //match at the last char
    idx=s1.rfind(s5);
    CPPUNIT_ASSERT(idx==43);
    CPPUNIT_ASSERT(s1[idx]=='g');
}

void
testSBuf::testRFindChar()
{
    SBuf s1(literal);
    SBuf::size_type idx;
    idx=s1.rfind(' ');
    CPPUNIT_ASSERT(idx == 40);
    CPPUNIT_ASSERT(s1[idx]==' ');
}

void
testSBuf::testRFindSBuf()
{
    SBuf s1(literal),s2("fox"),s3("goobar"),s4("foe"), s5("dog");
    //very long string matching at the end
    s5.append(literal).append(literal).append(literal).append(literal);

    SBuf::size_type idx;
    idx=s1.find(s2); //found
    CPPUNIT_ASSERT(idx == 16);
    idx=s3.find(s2); //fails, no initial char match
    CPPUNIT_ASSERT(idx==SBuf::npos);
    idx=s1.find(s4); //fails, no post-initial match
    CPPUNIT_ASSERT(idx==SBuf::npos);
    idx=s1.rfind(s5); //fails, too long after initial match
    CPPUNIT_ASSERT(idx==SBuf::npos);
    s5="g"; //match at the last char
    idx=s1.rfind(s5);
    CPPUNIT_ASSERT(idx==43);
    CPPUNIT_ASSERT(s1[idx]=='g');
}

void
testSBuf::testSBufLength()
{
    SBuf s(fox);
    CPPUNIT_ASSERT((size_t)s.length()==strlen(fox));
}

void
testSBuf::testScanf()
{
    SBuf s1;
    char s[128];
    int i;
    float f;
    int rv;
    s1.assign("string , 123 , 123.50");
    rv=s1.scanf("%s , %d , %f",s,&i,&f);
    CPPUNIT_ASSERT(3 == rv);
    CPPUNIT_ASSERT(0 == strcmp(s,"string"));
    CPPUNIT_ASSERT(i == 123);
    CPPUNIT_ASSERT(f == 123.5);
}

void testSBuf::testCopy()
{
    char buf[40]; //shorter than literal()
    SBuf s(fox1),s2;
    CPPUNIT_ASSERT(s.copy(buf,40)==s.length());
    CPPUNIT_ASSERT(strncmp(s.rawContent(),buf,s.length())==0);
    s=literal;
    CPPUNIT_ASSERT(s.copy(buf,40)==40);
    s2.assign(buf,0,40);
    s.chop(0,40);
    CPPUNIT_ASSERT(s==s2);
}

// XXX: please split each class into a separate unit-test .cc / even if they share a binary.
//  rule-of-thumb guideline for layering is 'one class Foo -> one Foo.cc,  one Foo.cc -> one testFoo.cc'

static int sbuf_tokens_number=9;
static SBuf tokens[]={
    SBuf("The",0,3), SBuf("quick",0,5), SBuf("brown",0,5), SBuf("fox",0,3),
    SBuf("jumped",0,6), SBuf("over",0,4), SBuf("the",0,3), SBuf("lazy",0,4),
    SBuf("dog",0,3)
};

void testSBuf::testSBufTokenizer()
{
    int j=0;
    SBuf s;
    for (SBufTokenizer st(literal,SBuf(" ",0,1)); !st.atEnd(); st.next()) {
        s=st.token();
        CPPUNIT_ASSERT(s==tokens[j]);
        j++;
    }
    CPPUNIT_ASSERT(j==9);
}

void testSBuf::testStringOps()
{
    SBuf sng(literal),
    ref("the quick brown fox jumped over the lazy dog");
    sng=sng.toLower();
    CPPUNIT_ASSERT_EQUAL(ref,sng);
    sng=literal;
    CPPUNIT_ASSERT(0==sng.compare(ref,caseInsensitive));
}

void testSBuf::testGrow()
{
    SBuf t;
    t.assign("foo");
    const char *ref=t.rawContent();
    t.reserve(10240);
    const char *match=t.rawContent();
    CPPUNIT_ASSERT(match!=ref);
    ref=match;
    t.append(literal).append(literal).append(literal).append(literal).append(literal);
    t.append(t).append(t).append(t).append(t).append(t);
    CPPUNIT_ASSERT(match==ref);
}

void testSBuf::testStartsWith()
{
    static SBuf casebuf("THE QUICK");
    CPPUNIT_ASSERT(literal.startsWith(SBuf(fox1)));
    CPPUNIT_ASSERT(!SBuf("The quick brown").startsWith(SBuf(fox1))); //too short
    CPPUNIT_ASSERT(!literal.startsWith(SBuf(fox2))); //wrong contents

    CPPUNIT_ASSERT(literal.startsWith(casebuf,caseInsensitive));
    casebuf=SBuf(fox1).toUpper();
    CPPUNIT_ASSERT(literal.startsWith(casebuf,caseInsensitive));
    CPPUNIT_ASSERT(literal.startsWith(SBuf(fox1),caseInsensitive));
    casebuf = "tha quick";
    CPPUNIT_ASSERT(!literal.startsWith(casebuf,caseInsensitive));
}

void testSBuf::testSBufList()
{
    SBufList foo;
    for (int j=0; j<sbuf_tokens_number; ++j)
        foo.add(tokens[j]);
    CPPUNIT_ASSERT(foo.isMember(SBuf("fox")));
    CPPUNIT_ASSERT(foo.isMember(SBuf("Fox"),caseInsensitive));
    CPPUNIT_ASSERT(!foo.isMember(SBuf("garble")));
    CPPUNIT_ASSERT(foo.isPrefix(SBuf("qui")));
    CPPUNIT_ASSERT(foo.isPrefix(SBuf("qUi"),caseInsensitive));
    CPPUNIT_ASSERT(!foo.isPrefix(SBuf("qUa"),caseInsensitive));
}

void testSBuf::testBaseName()
{
    SBuf totest("/foo/bar/gazonk");
    CPPUNIT_ASSERT_EQUAL(BaseName(totest),SBuf("gazonk"));
    CPPUNIT_ASSERT_EQUAL(BaseName(totest,'.'),totest);
}

void testSBuf::testSBufStream()
{
    SBuf b("const.string, int 10 and a float 10.5");
    SBufStream ss;
    ss << "const.string, int " << 10 << " and a float " << 10.5;
    SBuf o=ss.buf();
    CPPUNIT_ASSERT_EQUAL(b,o);
    ss.clearBuf();
    o=ss.buf();
    CPPUNIT_ASSERT_EQUAL(SBuf(),o);
    SBuf f1(fox1);
    SBufStream ss2(f1);
    ss2 << fox2;
    CPPUNIT_ASSERT_EQUAL(ss2.buf(),literal);
    CPPUNIT_ASSERT_EQUAL(f1,SBuf(fox1));
}
